from dataclasses import dataclass, field
from joserfc.errors import InvalidClaimError, MissingClaimError
from joserfc.jwt import JWTClaimsRegistry
from logging import DEBUG, getLogger
from re import compile, error, IGNORECASE, NOFLAG, Pattern, RegexFlag
from starlette.types import ASGIApp, Receive, Scope, Send
from .constants import (
    _DEFAULT_ANY_HTTP_METHODS,
    _DEFAULT_CLAIMS_CALLABLE,
    _DEFAULT_RAISE_ON_INVALID_CLAIM,
    _DEFAULT_RAISE_ON_INVALID_CLAIMS_TYPE,
    _DEFAULT_RAISE_ON_MISSING_CLAIM,
    _DEFAULT_RAISE_ON_UNAUTHENTICATED,
    _DEFAULT_RAISE_ON_UNSPECIFIED_METHOD,
    _DEFAULT_RAISE_ON_UNSPECIFIED_PATH,
    _DEFAULT_RE_IGNORECASE,
    _DEFAULT_SECURED,
    _DEFAULT_SKIPPED,
)
from .decorators import (
    validate_claims_callable,
    validate_secured,
    validate_skipped,
)
from .exceptions import (
    InvalidClaimsTypeException,
    InvalidClaimValueException,
    MissingEssentialClaimException,
    UnauthenticatedRequestException,
    UnspecifiedMethodAuthenticationException,
    UnspecifiedPathAuthenticationException,
)
from .types import (
    ClaimsCallableType, 
    SecuredCompiledType, 
    SecuredType, 
    SkippedCompiledType, 
    SkippedType,
)

log = getLogger(__name__)

@dataclass
class ClaimValidatorMiddleware:
    """
    Middleware for validating JWT claims in ASGI applications.

    Attributes:
        app (ASGIApp): The ASGI application.
        claims_callable (ClaimsCallableType): A callable that receives the current call scope and returns the claims as a dict.
        raise_on_invalid_claim (bool): Flag to raise an exception on invalid claims.
        raise_on_invalid_claims_type (bool): Flag to raise an exception on invalid claims type.
        raise_on_missing_claim (bool): Flag to raise an exception on missing claims.
        raise_on_unauthenticated (bool): Flag to raise an exception on unauthenticated requests.
        raise_on_unspecified_method (bool): Flag to raise an exception on unspecified methods.
        raise_on_unspecified_path (bool): Flag to raise an exception on unspecified paths.
        re_flags (RegexFlag): Regular expression flags.
        re_ignorecase (bool): Flag to ignore case in regular expressions.
        secured_compiled (SecuredCompiledType): Compiled secured paths and methods.
        secured (SecuredType): Secured paths and methods.
        skipped_compiled (SkippedCompiledType): Compiled skipped paths and methods.
        skipped (SkippedType): Skipped paths and methods.
    """
    app: ASGIApp
    claims_callable: ClaimsCallableType = field(default=_DEFAULT_CLAIMS_CALLABLE)
    raise_on_invalid_claim: bool = field(default=_DEFAULT_RAISE_ON_INVALID_CLAIM)
    raise_on_invalid_claims_type: bool = field(default=_DEFAULT_RAISE_ON_INVALID_CLAIMS_TYPE)
    raise_on_missing_claim: bool = field(default=_DEFAULT_RAISE_ON_MISSING_CLAIM)
    raise_on_unauthenticated: bool = field(default=_DEFAULT_RAISE_ON_UNAUTHENTICATED)
    raise_on_unspecified_method: bool = field(default=_DEFAULT_RAISE_ON_UNSPECIFIED_METHOD)
    raise_on_unspecified_path: bool = field(default=_DEFAULT_RAISE_ON_UNSPECIFIED_PATH)
    re_flags: RegexFlag = field(default=NOFLAG, init=False)
    re_ignorecase: bool = field(default=_DEFAULT_RE_IGNORECASE)
    secured_compiled: SecuredCompiledType = field(default_factory=dict, init=False)
    secured: SecuredType = field(default_factory=lambda: _DEFAULT_SECURED)
    skipped_compiled: SkippedCompiledType = field(default_factory=dict, init=False)
    skipped: SkippedType = field(default_factory=lambda: _DEFAULT_SKIPPED)
  
    @validate_claims_callable()
    @validate_secured()
    @validate_skipped()
    def __post_init__(self) -> None:
        """
        Post-initialization method to compile regular expressions for secured and skipped paths.

        This method is called after the object is initialized. It compiles the regular expressions
        for the paths specified in the `secured` and `skipped` attributes, and associates them with
        their corresponding HTTP methods and claims.

        Attributes:
            re_flags (int): Regular expression flags, set to IGNORECASE if `re_ignorecase` is True, otherwise NOFLAG.
            secured_compiled (dict): A dictionary where keys are compiled regular expressions for secured paths,
                                    and values are dictionaries mapping HTTP methods to their corresponding claims.
            skipped_compiled (dict): A dictionary where keys are compiled regular expressions for skipped paths,
                                    and values are sets of HTTP methods in uppercase.

        Raises:
            ValueError: If there is an invalid regular expression in the `secured` or `skipped` paths.
        """
        try:
            self.re_flags = IGNORECASE if self.re_ignorecase else NOFLAG
            # This code compiles regular expressions for each path in self.secured and associates them with a 
            # dictionary of HTTP methods in uppercase and their corresponding claims.
            self.secured_compiled = {
                compile(path, flags=self.re_flags): { 
                    method.upper(): claims for method, claims in methods.items()
                } for path, methods in self.secured.items()
            }
            # This code compiles regular expressions for each path in self.skipped and associates them with the 
            # corresponding HTTP methods in uppercase.
            self.skipped_compiled = {
                compile(path, flags=self.re_flags): set(
                    method.upper() for method in methods
                ) for path, methods in self.skipped.items()
            }
        except error as e:
            raise ValueError(f"Invalid regular expression in secured or skipped paths: {e}")

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        ASGI application callable to validate JWT claims.

        Args:
            scope (Scope): The ASGI scope.
            receive (Receive): The ASGI receive callable.
            send (Send): The ASGI send callable.

        Raises:
            InvalidClaimValueException: If a claim value is invalid and raise_on_invalid_claim is True.
            MissingEssentialClaimException: If a required claim is missing and raise_on_missing_claim is True.
            UnauthenticatedRequestException: If the request is unauthenticated and raise_on_unauthenticated is True.
            UnspecifiedMethodAuthenticationException: If the method is not specified and raise_on_unspecified_method is True.
            UnspecifiedPathAuthenticationException: If the path is not specified and raise_on_unspecified_path is True.
        """
        if scope["type"] not in ("http",):
            await self.app(scope, receive, send)
            return

        method = scope["method"].upper()
        path = scope["path"]
        claims = self.claims_callable(scope)

        # Check if the request path matches any skipped path patterns and if the request method is allowed for that path.
        # If both conditions are met, forward the request to the next middleware or application.
        for p in self._search_patterns_in_string(path, self.skipped_compiled.keys()):
            if any(sc_method in (method, _DEFAULT_ANY_HTTP_METHODS) for sc_method in self.skipped_compiled[p]):
                await self.app(scope, receive, send)
                return

        if self.raise_on_invalid_claims_type and not isinstance(claims, dict):
            raise InvalidClaimsTypeException(path=path, method=method, type_received=type(claims), type_expected=dict)

        if self.raise_on_unauthenticated and not claims:
            raise UnauthenticatedRequestException(path=path, method=method)

        # This dictionary comprehension filters the secured_compiled dictionary to include only those patterns
        # that match the current URL path. It creates a new dictionary where the keys are the patterns that match the URL path.
        filtered_patterns = {
            p: self.secured_compiled[p] for p in self._search_patterns_in_string(path, self.secured_compiled.keys())    
        }

        if self.raise_on_unspecified_path and not filtered_patterns:
            raise UnspecifiedPathAuthenticationException(path=path, method=method)

        # This dictionary comprehension filters the previously created filtered_patterns dictionary to include only those methods
        # that match the current HTTP method. It creates a new dictionary where the keys are the paths and the values
        # are dictionaries of methods and their corresponding claims that match the current HTTP method.
        filtered_patterns = {
            fp_path: {
                fp_method: fp_claims for fp_method, fp_claims in fp_methods.items() if fp_method in (method, _DEFAULT_ANY_HTTP_METHODS)
            } for fp_path, fp_methods in filtered_patterns.items()
        }

        if self.raise_on_unspecified_method and all(not fp_methods for fp_methods in filtered_patterns.values()):
            raise UnspecifiedMethodAuthenticationException(path=path, method=method)

        # This block iterates over filtered patterns of paths and methods, validates JWT claims for each method,
        # and logs any errors encountered during the validation process.
        for fp_path, fp_methods in filtered_patterns.items():
            for fp_method, fp_claims in fp_methods.items():
                if log.isEnabledFor(DEBUG):
                    log.debug(f"path: {path} | method: {method} | claims: {claims}")
                    log.debug(f"fp_path: {fp_path} | fp_method: {fp_method} | fp_claims: {fp_claims}")
                try:
                    claims_requests = JWTClaimsRegistry(**fp_claims)
                    claims_requests.validate(claims)
                except MissingClaimError as e:
                    log.debug(f"Missing claim: {e}")
                    if self.raise_on_missing_claim:
                        raise MissingEssentialClaimException(path=path, method=method, claims=claims)
                except InvalidClaimError as e:
                    log.debug(f"Invalid claim: {e}")
                    if self.raise_on_invalid_claim:
                        raise InvalidClaimValueException(path=path, method=method, claims=claims)
                except Exception as e:
                    log.error(f"Unexpected error during claim validation: {e}")
                    raise e

        await self.app(scope, receive, send)
        return

    @staticmethod
    def _search_patterns_in_string(s: str, patterns: list[Pattern]) -> list[Pattern]:
        """
        Searches for patterns in a given string and returns the patterns that match.

        This method iterates over a list of compiled regular expression patterns and checks if each pattern matches the given string `s`. 
        The method returns a list of patterns that have at least one match in the string.

        Args:
            s (str): The string to search within.
            patterns (list[Pattern]): A list of compiled regular expression patterns to search for.

        Returns:
            list[Pattern]: A list of patterns that match the string.
        """
        return [p for p in patterns if p.search(s)]


from re import escape
from .types import SecuredType, SkippedType, ClaimsCallableType

_DEFAULT_ANY_HTTP_METHODS: str = "*"
_DEFAULT_ALL_HTTP_METHODS: list[str] = [
    "CONNECT", 
    "DELETE", 
    "GET", 
    "HEAD", 
    "OPTIONS", 
    "PATCH", 
    "POST", 
    "PUT", 
    "TRACE",
]
_DEFAULT_ALL_HTTP_METHODS_REGEX_GROUP: str = f"({'|'.join(map(escape, (*_DEFAULT_ALL_HTTP_METHODS, *_DEFAULT_ANY_HTTP_METHODS)))})"
_DEFAULT_CLAIMS_CALLABLE: ClaimsCallableType = lambda scope: scope.get("", dict())
_DEFAULT_RAISE_ON_INVALID_CLAIM: bool = True
_DEFAULT_RAISE_ON_INVALID_CLAIMS_TYPE: bool = True
_DEFAULT_RAISE_ON_MISSING_CLAIM: bool = True
_DEFAULT_RAISE_ON_UNAUTHENTICATED: bool = True
_DEFAULT_RAISE_ON_UNSPECIFIED_METHOD: bool = True
_DEFAULT_RAISE_ON_UNSPECIFIED_PATH: bool = True
_DEFAULT_RE_IGNORECASE: bool = False
_DEFAULT_SECURED: SecuredType = {
    "^$": {
        f"{_DEFAULT_ANY_HTTP_METHODS}": {
            "sub": {
                "essential": False, 
                "allow_blank": False,
            },
        },
    },
}
_DEFAULT_SKIPPED: SkippedType = {
    "^$": [
        f"{_DEFAULT_ANY_HTTP_METHODS}",
    ],
}
_DEFAULT_SKIPPED_JSON_SCHEMA: dict = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title": "Skipped JSON Schema",
    "description": "Schema for validating skipped configuration",
    "type": "object",
    "minProperties": 1,
    "patternProperties": {
        "^(.+)$": {
            "type": "array",
            "items": {
                "oneOf": [
                    {
                        "type": "string",
                        "pattern": f"(?i:(^({_DEFAULT_ALL_HTTP_METHODS_REGEX_GROUP})$))",
                    },
                    { 
                        "type": "null",
                    },
                ],
            },
        },
    },
    "additionalProperties": False,
    "unevaluatedProperties": False,
}
_DEFAULT_SECURED_JSON_SCHEMA: dict = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title": "Secured JSON Schema",
    "description": "Schema for validating secured configuration",
    "type": "object",
    "minProperties": 1,
    "patternProperties": {
        "(^(.+)$)": {
            "type": "object",
            "minProperties": 1,
            "patternProperties": {
                f"(?i:(^({_DEFAULT_ALL_HTTP_METHODS_REGEX_GROUP})$))": {
                    "type": "object",
                    "minProperties": 1,
                    "patternProperties": {
                        "(^(.+)$)": {
                            "type": "object",
                            "minProperties": 1,
                            "properties": {
                                "essential": {
                                    "oneOf": [
                                        {
                                            "type": "boolean",
                                        },
                                    ],
                                },
                                "allow_blank": {
                                    "oneOf": [
                                        {
                                            "type": "boolean",
                                        },
                                        {
                                            "type": "null",
                                        },
                                    ],
                                },
                                "values": {
                                    "type": "array",
                                    "items": {
                                        "oneOf": [
                                            {
                                                "type": "boolean",
                                            },
                                            {
                                                "type": "integer",
                                            },
                                            {
                                                "type": "string",
                                            },
                                        ],
                                    },
                                },
                            },
                            "additionalProperties": False,
                            "unevaluatedProperties": False,
                        },
                    },
                    "additionalProperties": False,
                    "unevaluatedProperties": False,
                },
            },
            "additionalProperties": False,
            "unevaluatedProperties": False,
        },
    },
    "additionalProperties": False,
    "unevaluatedProperties": False,
}


from collections.abc import Callable
from joserfc.jwt import ClaimsOption, Claims
from re import Pattern
from starlette.types import Scope

SecuredCompiledType = dict[Pattern, dict[str, dict[str, ClaimsOption]]]
SecuredType = dict[str, dict[str, dict[str, ClaimsOption]]]
SkippedCompiledType = dict[Pattern, set[str]]
SkippedType = dict[str, list[str]]
ClaimsType = Claims
ClaimsCallableType = Callable[[Scope], Claims]


class ClaimValidatorException(Exception):
    """Base exception for claim validator-related errors.

    This is the base class for all exceptions raised by the claim validator.
    It provides a common interface for handling errors related to claim validation.

    Attributes:
        detail (str): A detailed error message.
        status (int): The HTTP status code.
        title (str): A short HTTP status message.
    """
    description: str = "A claim validator error occurred."
    status: int = 400
    title: str = "Bad Request"

    def __init__(self, detail: str = description, status: int = status, title: str = title) -> None:
        self.detail: str = detail
        self.status: int = status
        self.title: str = title
        super().__init__(self.detail)

class UnspecifiedMethodAuthenticationException(ClaimValidatorException):
    """Exception raised when authentication is not specified for a method.

    This exception is used in the ClaimValidatorMiddleware to indicate that 
    the specified HTTP method does not have an associated authentication configuration.
    It is raised when the `raise_on_unspecified_method` flag is set to True and 
    no matching secured method pattern is found for the current request method.

    Attributes:
        method (str): The HTTP method of the request.
        path (str): The path of the request.
        detail (str): A detailed error message.
        status (int): The HTTP status code.
        title (str): A short HTTP status message.
    """
    description: str = (
        "Authentication configuration missing for the specified method. "
        "Ensure that an appropriate authentication definition is provided "
        "for this method and try again."
    )
    status: int = 401
    title: str = "Unauthorized"

    def __init__(self, method: str, path: str, detail: str = description, status: int = status, title: str = title) -> None:
        self.method: str = method
        self.path: str = path
        self.detail: str = detail
        self.status: int = status
        self.title: str = title
        super().__init__(self.detail, self.status, self.title)

    def __str__(self) -> str:
        return f"Method authentication not specified {self.method} {self.path} ({self.detail})"

class UnspecifiedPathAuthenticationException(ClaimValidatorException):
    """Exception raised when authentication is not specified for a path.

    This exception is used in the ClaimValidatorMiddleware to indicate that 
    the specified path does not have an associated authentication configuration.
    It is raised when the `raise_on_unspecified_path` flag is set to True and 
    no matching secured path pattern is found for the current request path.

    Attributes:
        method (str): The HTTP method of the request.
        path (str): The path of the request.
        detail (str): A detailed error message.
        status (int): The HTTP status code.
        title (str): A short HTTP status message.
    """
    description: str = (
        "Authentication configuration missing for the specified path. "
        "Ensure that an appropriate authentication definition is provided "
        "for this path and try again."
    )
    status: int = 401
    title: str = "Unauthorized"

    def __init__(self, method: str, path: str, detail: str = description, status: int = status, title: str = title) -> None:
        self.method: str = method
        self.path: str = path
        self.detail: str = detail
        self.status: int = status
        self.title: str = title
        super().__init__(self.detail, self.status, self.title)

    def __str__(self) -> str:
        return f"Path authentication not specified {self.method} {self.path} ({self.detail})"

class UnauthenticatedRequestException(ClaimValidatorException):
    """Exception raised when a request cannot be authenticated.

    This exception is used in the ClaimValidatorMiddleware to indicate that 
    the request could not be authenticated due to missing or invalid claims.
    It is raised when the `raise_on_unauthenticated` flag is set to True and 
    the claims provided are insufficient for authentication.

    Attributes:
        path (str): The path of the request.
        method (str): The HTTP method of the request.
        detail (str): A detailed error message.
        status (int): The HTTP status code.
        title (str): A short HTTP status message.
    """
    description: str = (
        "The request could not be authenticated. Ensure that the necessary "
        "claims are provided and try again."
    )
    status: int = 401
    title: str = "Unauthorized"

    def __init__(self, path: str, method: str, detail: str = description, status: int = status, title: str = title) -> None:
        self.path: str = path
        self.method: str = method
        self.detail: str = detail
        self.status: int = status
        self.title: str = title
        super().__init__(self.detail, self.status, self.title)

    def __str__(self) -> str:
        return f"Unauthenticated request {self.method} {self.path} ({self.detail})"

class MissingEssentialClaimException(ClaimValidatorException):
    """Exception raised when an essential claim is missing from the request.

    This exception is used in the ClaimValidatorMiddleware to indicate that 
    a required claim is missing from the JWT claims provided in the request.
    It is raised when the `raise_on_missing_claim` flag is set to True and 
    a required claim is not found in the JWT claims.

    Attributes:
        path (str): The path of the request.
        method (str): The HTTP method of the request.
        claims (str): The claims in the request.
        detail (str): A detailed error message.
        status (int): The HTTP status code.
        title (str): A short HTTP status message.
    """
    description: str = (
        "An essential claim is missing from the request. Ensure that the "
        "necessary claims are provided and try again."
    )
    status: int = 403
    title: str = "Forbidden"

    def __init__(self, path: str, method: str, claims: str, detail: str = description, status: int = status, title: str = title) -> None:
        self.path: str = path
        self.method: str = method
        self.claims: str = claims
        self.detail: str = detail
        self.status: int = status
        self.title: str = title
        super().__init__(self.detail, self.status, self.title)

    def __str__(self) -> str:
        return f"Missing essential claims in request {self.method} {self.path} {self.claims} ({self.detail})"

class InvalidClaimValueException(ClaimValidatorException):
    """Exception raised when a claim has an invalid value.

    This exception is used in the ClaimValidatorMiddleware to indicate that 
    a claim provided in the JWT claims has an invalid value.
    It is raised when the `raise_on_invalid_claim` flag is set to True and 
    a claim is found to have an invalid value during validation.

    Attributes:
        path (str): The path of the request.
        method (str): The HTTP method of the request.
        claims (str): The claims in the request.
        detail (str): A detailed error message.
        status (int): The HTTP status code.
        title (str): A short HTTP status message.
    """
    description: str = (
        "A claim has an invalid value. Ensure that the claims provided have "
        "valid values and try again."
    )
    status: int = 403
    title: str = "Forbidden"

    def __init__(self, path: str, method: str, claims: str, detail: str = description, status: int = status, title: str = title) -> None:
        self.path: str = path
        self.method: str = method
        self.claims: str = claims
        self.detail: str = detail
        self.status: int = status
        self.title: str = title
        super().__init__(self.detail, self.status, self.title)

    def __str__(self) -> str:
        return f"Invalid claims value in request {self.method} {self.path} {self.claims} ({self.detail})"
    
class InvalidClaimsTypeException(ClaimValidatorException):
    """Exception raised when the claims provided are not of the expected type.

    This exception is raised when the claims provided to the ClaimValidatorMiddleware
    are not of the expected type. It indicates that the claims should be a dictionary
    but are not.

    Attributes:
        path (str): The path of the request.
        method (str): The HTTP method of the request.
        type_received (str): The type of the claims received.
        type_expected (str): The expected type of the claims.
        detail (str): A detailed error message.
        status (int): The HTTP status code.
        title (str): A short HTTP status message.
    """
    description: str = (
        "The claims provided are not of the expected type. Ensure that the claims are "
        "correctly formatted as a dictionary and try again."
    )
    status: int = 400
    title: str = "Bad Request"

    def __init__(self, path: str, method: str, type_received: str, type_expected: str, detail: str = description, status: int = status, title: str = title) -> None:
        self.path: str = path
        self.method: str = method
        self.type_received: str = type_received
        self.type_expected: str = type_expected
        self.detail: str = detail
        self.status: int = status
        self.title: str = title
        super().__init__(self.detail, self.status, self.title)

    def __str__(self) -> str:
        return f"Invalid claims type in request {self.method} {self.path} (received: {self.type_received}; expected: {self.type_expected}) ({self.detail})"

class InvalidClaimsConfigurationException(ClaimValidatorException):
    """Exception raised when the claims configuration is invalid.

    This exception is used to indicate that the claims callable provided
    does not return an instance of Claims. It is raised during the initialization
    of the ClaimValidatorMiddleware when the `claims` parameter is not callable
    or does not return the expected type.

    Attributes:
        detail (str): A detailed error message.
        status (int): The HTTP status code.
        title (str): A short HTTP status message.
    """
    description: str = (
        "The claims callable must return an instance of Claims. Ensure that the "
        "claims callable is correctly referenced and returns the expected type."
    )
    status: int = 500
    title: str = "Internal Server Error"

    def __init__(self, detail: str = description, status: int = status, title: str = title) -> None:
        self.detail: str = detail
        self.status: int = status
        self.title: str = title
        super().__init__(self.detail, self.status, self.title)

    def __str__(self) -> str:
        return f"Invalid claims callable: {self.detail}"

class InvalidSecuredConfigurationException(ClaimValidatorException):
    """Exception raised when the secured configuration is invalid.

    This exception is used to indicate that the secured dictionary provided
    is not correctly formatted or contains invalid values. It is raised during
    the initialization of the ClaimValidatorMiddleware.

    Attributes:
        detail (str): A detailed error message.
        status (int): The HTTP status code.
        title (str): A short HTTP status message.
    """
    description: str = (
        "The secured configuration is invalid. Ensure that the secured dictionary "
        "is correctly formatted and contains valid values."
    )
    status: int = 500
    title: str = "Internal Server Error"

    def __init__(self, detail: str = description, status: int = status, title: str = title) -> None:
        self.detail: str = detail
        self.status: int = status
        self.title: str = title
        super().__init__(self.detail, self.status, self.title)

    def __str__(self) -> str:
        return f"Invalid secured configuration: {self.detail}"

class InvalidSkippedConfigurationException(ClaimValidatorException):
    """Exception raised when the skipped configuration is invalid.

    This exception is used to indicate that the skipped dictionary provided
    is not correctly formatted or contains invalid values. It is raised during
    the initialization of the ClaimValidatorMiddleware.

    Attributes:
        detail (str): A detailed error message.
        status (int): The HTTP status code.
        title (str): A short HTTP status message.
    """
    description: str = (
        "The skipped configuration is invalid. Ensure that the skipped dictionary "
        "is correctly formatted and contains valid values."
    )
    status: int = 500
    title: str = "Internal Server Error"

    def __init__(self, detail: str = description, status: int = status, title: str = title) -> None:
        self.detail: str = detail
        self.status: int = status
        self.title: str = title
        super().__init__(self.detail, self.status, self.title)

    def __str__(self) -> str:
        return f"Invalid skipped configuration: {self.detail}"


from collections.abc import Callable
from jsonschema import validate
from jsonschema.exceptions import SchemaError, ValidationError
from logging import getLogger
from .constants import (
    _DEFAULT_CLAIMS_CALLABLE, 
    _DEFAULT_SECURED_JSON_SCHEMA,
    _DEFAULT_SKIPPED_JSON_SCHEMA,
)
from .exceptions import (
    InvalidClaimsConfigurationException, 
    InvalidSecuredConfigurationException,
    InvalidSkippedConfigurationException,
)

log = getLogger(__name__)

def validate_claims_callable() -> Callable:
    """
    Decorator to validate the claims_callable attribute of a class.

    Raises:
        InvalidClaimsConfigurationException: If claims_callable is not a callable.
    """
    def decorator(func: Callable) -> Callable:
        def wrapper(self, *args, **kwargs) -> Callable:
            claims = getattr(self, 'claims_callable', _DEFAULT_CLAIMS_CALLABLE)
            if not isinstance(claims, Callable):
                raise InvalidClaimsConfigurationException()
            return func(self, *args, **kwargs)
        return wrapper
    return decorator

def validate_secured() -> Callable:
    """
    Decorator to validate the secured attribute of a class against a JSON schema.

    Raises:
        InvalidSecuredConfigurationException: If the secured attribute does not conform to the schema.
    """
    def decorator(func: Callable) -> Callable:
        def wrapper(self, *args, **kwargs) -> Callable:
            secured = getattr(self, 'secured', None)
            try:
                validate(instance=secured, schema=_DEFAULT_SECURED_JSON_SCHEMA)
            except (SchemaError, ValidationError) as e:
                log.error(e)
                raise InvalidSecuredConfigurationException()
            return func(self, *args, **kwargs)
        return wrapper
    return decorator

def validate_skipped() -> Callable:
    """
    Decorator to validate the skipped attribute of a class against a JSON schema.

    Raises:
        InvalidSkippedConfigurationException: If the skipped attribute does not conform to the schema.
    """
    def decorator(func: Callable) -> Callable:
        def wrapper(self, *args, **kwargs) -> Callable:
            skipped = getattr(self, 'skipped', None)
            try:
                validate(instance=skipped, schema=_DEFAULT_SKIPPED_JSON_SCHEMA)
            except (SchemaError, ValidationError) as e:
                log.error(e)
                raise InvalidSkippedConfigurationException()
            return func(self, *args, **kwargs)
        return wrapper
    return decorator


from .exceptions import (
    ClaimValidatorException,
    InvalidClaimsConfigurationException,
    InvalidClaimsTypeException,
    InvalidClaimValueException,
    InvalidSecuredConfigurationException,
    InvalidSkippedConfigurationException,
    MissingEssentialClaimException,
    UnauthenticatedRequestException,
    UnspecifiedMethodAuthenticationException,
    UnspecifiedPathAuthenticationException,
)
from .middleware import ClaimValidatorMiddleware
from .types import (
    ClaimsCallableType,
    ClaimsType,
    SecuredCompiledType,
    SecuredType,
    SkippedCompiledType,
    SkippedType,
)

__all__ = (
    "ClaimsCallableType",
    "ClaimsType",
    "ClaimValidatorException",
    "ClaimValidatorMiddleware",
    "InvalidClaimsConfigurationException",
    "InvalidClaimsTypeException",
    "InvalidClaimValueException",
    "InvalidSecuredConfigurationException",
    "InvalidSkippedConfigurationException",
    "MissingEssentialClaimException",
    "SecuredCompiledType",
    "SecuredType",
    "SkippedCompiledType",
    "SkippedType",
    "UnauthenticatedRequestException",
    "UnspecifiedMethodAuthenticationException",
    "UnspecifiedPathAuthenticationException",
)